# coding: utf-8
"""
MPI Tests for MPI Processor
"""
import sys
import logging
import numpy as np
from maggma.builder import Builder
from maggma.stores import MemoryStore
from maggma.runner import MPIProcessor


class DummyBuilder(Builder):

    def __init__(self, temp_storage_store):
        self.temp_storage_store = temp_storage_store
        super(DummyBuilder, self).__init__(sources=[], targets=[temp_storage_store],chunk_size=100)

    def get_items(self):
        self.logger.info("Getting Items")
        for i in range(1000):
            yield {"val": i, "task_id": i}

    def process_item(self, item):
        if item["val"] % 10 == 0:
            self.logger.debug("Processing: {}".format(item["val"]))
        proc_val = np.sqrt(np.square(float(item["val"])))
        item["proc_val"] = proc_val
        return item

    def update_targets(self, items):
        self.logger.info("Updating {} items".format(len(items)))
        self.temp_storage_store.update(items)

if __name__ == "__main__":

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    ch = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    ch.setFormatter(formatter)
    root.addHandler(ch)

    mem = MemoryStore("processed")
    bldr = DummyBuilder(mem)

    mpi_proc = MPIProcessor([bldr])
    mpi_proc.process(0)



# coding: utf-8
"""
Module defining objects to run builders in various modes
including serial processing, multiprocessing on a single computer,
and processing via MPI
"""

import abc
import logging
import types
from collections import defaultdict, deque
from threading import Thread, Condition, BoundedSemaphore
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from monty.json import MSONable
from maggma.utils import get_mpi, grouper
from tqdm import tqdm


class BaseProcessor(MSONable, metaclass=abc.ABCMeta):
    """
    Base processor class for multiprocessing paradigms
    """

    def __init__(self, builders):
        """
        Initialize with a list of builders

        Args:
            builders(list): list of builders
        """
        self.builders = builders

        self.logger = logging.getLogger(type(self).__name__)
        self.logger.addHandler(logging.NullHandler())

    @abc.abstractmethod
    def process(self, builder_id):
        """
        Does the processing. e.g. send work to workers(in MPI) or start the processes in
        multiprocessing.

        Args:
            builder_id (int): process the builder_id th builder i.e
                process_item --> update_targets --> finalize
        """
        pass


class SerialProcessor(BaseProcessor):
    """
    Simple serial processor. Usefull for debugging or example code
    """

    def process(self, builder_id):
        """
        Run the builder serially

        Args:
            builder_id (int): the index of the builder in the builders list
        """
        builder = self.builders[builder_id]
        chunk_size = builder.chunk_size

        # establish connection to the sources and targets
        builder.connect()

        cursor = builder.get_items()

        for chunk in grouper(cursor, chunk_size):
            self.logger.info("Processing batch of {} items".format(chunk_size))
            processed_items = [builder.process_item(item) for item in chunk if item is not None]
            builder.update_targets(processed_items)


class MPIProcessor(BaseProcessor):
    """
    Processor to distribute work using MPI
    """

    def __init__(self, builders):
        (self.comm, self.rank, self.size) = get_mpi()
        if not self.comm:
            raise Exception(
                "MPI not working properly, check your mpi4py installation and ensure this is running under mpi")
        self.comm.barrier()
        super(MPIProcessor, self).__init__(builders)

    def process(self, builder_id):
        """
        Run the builder using MPI protocol.

        Args:
            builder_id (int): the index of the builder in the builders list
        """
        self.comm.barrier()
        if self.rank == 0:
            self.process_master(builder_id)
        else:
            self.process_worker()

    def setup_multithreading(self):
        """
        Setup structures for managing data to/from MPI Workers
        """
        self.data = deque()
        self.ranks = deque([i + 1 for i in range(self.size - 1)])
        self.task_count = BoundedSemaphore(self.builder.chunk_size)
        self.update_data_condition = Condition()

        self.run_update_targets = True
        self.update_targets_thread = Thread(target=self.update_targets)
        self.update_targets_thread.start()

    def process_master(self, builder_id):
        """
        Master process for MPI processing
        Handles Data IO to Stores and to MPI Workers
        """
        self.builder = self.builders[builder_id]
        self.builder.connect()

        cursor = self.builder.get_items()

        self.setup_pbars(cursor)
        self.setup_multithreading()
        self.put_tasks(builder_id)
        self.clean_up_workers()
        self.clean_up_data()
        self.builder.finalize(cursor)
        self.cleanup_pbars()

    def process_worker(self):
        """
        MPI Worker process
        """
        is_valid = True

        while is_valid:
            packet = self.comm.recv(source=0)
            if packet["type"] == "process":
                builder_id = packet["builder_id"]
                data = packet["data"]
                try:
                    result = self.builders[builder_id].process_item(data)
                    self.comm.send({"type": "return", "return": result}, dest=0)
                except e:
                    self.comm.send({"type": "error", "error": e})
            elif packet["type"] == "shutdown":
                is_valid = False

    def setup_pbars(self, cursor):
        """
        Sets up progress bars
        """
        total = None
        if hasattr(cursor, "__len__"):
            total = len(cursor)
        elif hasattr(cursor, "count"):
            total = cursor.count()
        elif hasattr(self.builder, "total"):
            total = self.builder.total

        self.get_pbar = tqdm(cursor, desc="Get Items", total=total)
        self.process_pbar = tqdm(desc="Processing Item", total=total)
        self.update_pbar = tqdm(desc="Updating Targets", total=total)

    def cleanup_pbars(self):
        """
        Cleans up the TQDM bars
        """
        self.get_pbar.close()
        self.process_pbar.close()
        self.update_pbar.close()

    def put_tasks(self, builder_id):
        """
        Submit tasks from cursor to MPI workers
        """
        # 1.) Setup thread pool
        with ThreadPoolExecutor(max_workers=self.size - 1) as executor:
            # 2.) Loop over every item wrapped in a tqdm bar
            for item in self.get_pbar:
                # 3.) Limit total number of queued tasks using a semaphore
                self.task_count.acquire()
                # 4.) Submit the item to a worker
                f = executor.submit(self.submit_item, builder_id, item)

    def submit_item(self, builder_id, data):
        """
        Thread to submit an item to MPI Workers and get data back

        """

        # 1.) Find free rank and take it
        mpi_rank = self.ranks.pop()
        # 2.) Submit the job to that rank
        self.comm.send({"type": "process", "builder_id": builder_id, "data": data}, dest=mpi_rank)
        # 3.) Periodically poll for data back
        result = None
        while not result:
            packet = self.comm.recv(source=mpi_rank)
            if packet["type"] == "return":
                result = packet["return"]
                self.task_count.release()
            elif packet["type"] == "error":
                self.logger.error("MPI Rank {} Errored on Builder ID {}:\n{}".format(
                    mpi_rank, builder_id, packet["error"]))
                self.task_count.release()
                return
            else:
                self.task_count.release()
                return  # don't know what happened here, just quit

        # 6.) Update process progress bar
        self.process_pbar.update(1)

        # 7.) Save data
        with self.update_data_condition:
            self.data.append(result)
            self.update_data_condition.notify_all()
        # 8.) Return rank
        self.ranks.append(mpi_rank)

    def clean_up_workers(self):
        """
        Sends shutdown signal to all MPI workers
        """
        for i in range(self.size - 1):
            self.comm.send({"type": "shutdown"}, dest=i + 1)

    def clean_up_data(self):
        """
        Call back to add data into a list in thread safe manner and signal other threads to add more tasks or update_targets
        """
        self.logger.debug("Cleaning up data queue")
        try:
            with self.update_data_condition:
                self.run_update_targets = False
                self.update_data_condition.notify_all()
        except Exception as e:
            self.logger.debug("Problem in updating targets at end of builder run: {}".format(e))

        self.update_targets_thread.join()

    def update_targets(self):
        """
        Thread to update targets periodically
        """
        while self.run_update_targets:
            with self.update_data_condition:
                self.update_data_condition.wait_for(
                    lambda: not self.run_update_targets or len(self.data) > self.builder.chunk_size)
                try:
                    self.builder.update_targets(self.data)
                    self.update_pbar.update(len(self.data))
                    self.data.clear()
                except Exception as e:
                    self.logger.debug("Problem in updating targets in builder run: {}".format(e))


class MultiprocProcessor(BaseProcessor):
    """
    Processor to run builders using python multiprocessing
    """

    def __init__(self, builders, num_workers=None):
        # multiprocessing only if mpi is not used, no mixing
        self.num_workers = num_workers
        super(MultiprocProcessor, self).__init__(builders)
        self.logger.info("Building with multiprocessing, {} workers in the pool".format(self.num_workers))

    def process(self, builder_id):
        """
        Run the builder using the builtin multiprocessing.

        Args:
            builder_id (int): the index of the builder in the builders list
        """
        self.builder = self.builders[builder_id]
        self.builder.connect()

        cursor = self.builder.get_items()

        self.setup_pbars(cursor)

        self.setup_multithreading()
        self.put_tasks()
        self.clean_up_data()
        self.builder.finalize(cursor)
        self.cleanup_pbars()

    def setup_pbars(self, cursor):
        """
        Sets up progress bars
        """
        total = None
        if hasattr(cursor, "__len__"):
            total = len(cursor)
        elif hasattr(cursor, "count"):
            total = cursor.count()
        elif isinstance(cursor,types.GeneratorType):
            cursor.send(None)
            if hasattr(self.builder, "total"):
                total = self.builder.total

        self.get_pbar = tqdm(cursor, desc="Get Items", total=total)
        self.process_pbar = tqdm(desc="Processing Item", total=total)
        self.update_pbar = tqdm(desc="Updating Targets", total=total)

    def cleanup_pbars(self):
        """
        Cleans up the TQDM bars
        """
        self.get_pbar.close()
        self.process_pbar.close()
        self.update_pbar.close()

    def setup_multithreading(self):
        """
        Sets up objects necessary to store and synchronize data in multiprocessing
        """
        self.data = deque()
        self.task_count = BoundedSemaphore(self.builder.chunk_size)
        self.update_data_condition = Condition()

        self.run_update_targets = True
        self.update_targets_thread = Thread(target=self.update_targets)
        self.update_targets_thread.start()

    def put_tasks(self):
        """
        Processes all items from builder using a pool of processes
        """
        # 1.) setup a process pool
        with ProcessPoolExecutor(self.num_workers) as executor:
            # 2.) Loop over every item wrapped in a tqdm bar
            for item in self.get_pbar:
                # 3.) Limit total number of queues tasks using a semaphore
                self.task_count.acquire()
                # 4.) Submit a task to processing pool
                f = executor.submit(self.builder.process_item, item)
                # 5.) Add call back to update our data list
                f.add_done_callback(self.update_data_callback)

    def clean_up_data(self):
        """
        Updates targets with remaining data and then cleans up the data collection
        """
        try:
            with self.update_data_condition:
                self.run_update_targets = False
                self.update_data_condition.notify_all()
        except Exception as e:
            self.logger.debug("Problem in updating targets at end of builder run: {}".format(e))

        self.update_targets_thread.join()

    def update_data_callback(self, future):
        """
        Call back to add data into a list in thread safe manner and signal other threads to add more tasks or update_targets
        """
        with self.update_data_condition:
            self.process_pbar.update(1)
            self.data.append(future.result())
            self.update_data_condition.notify_all()

        self.task_count.release()

    def update_targets(self):
        """
        Thread to update targets periodically
        """

        while self.run_update_targets:
            with self.update_data_condition:
                self.update_data_condition.wait_for(
                    lambda: not self.run_update_targets or len(self.data) > self.builder.chunk_size)
                try:
                    if self.data is not None:
                        self.update_pbar.unpause()
                        self.builder.update_targets(self.data)
                        self.update_pbar.update(len(self.data))
                        self.data.clear()
                except Exception as e:
                    self.logger.debug("Problem in updating targets in builder run: {}".format(e))


class Runner(MSONable):

    def __init__(self, builders, num_workers=None):
        """
        Initialize with a list of builders

        Args:
            builders(list): list of builders
            num_workers (int): number of processes. Used only for multiprocessing.
                Will be automatically set to (number of cpus - 1) if set to 0.
            processor(BaseProcessor): set this if custom processor is needed(must
                subclass BaseProcessor though)
        """
        self.builders = builders
        self.num_workers = num_workers
        self.logger = logging.getLogger(type(self).__name__)
        self.logger.addHandler(logging.NullHandler())
        (_, mpi_rank, mpi_size) = get_mpi()
        if mpi_size > 1:
            self.logger.info("Running with MPI Rank: {}".format(mpi_rank))
            self.processor = MPIProcessor(builders)
        else:
            self.logger.info("Running with Multiprocessing")
            self.processor = MultiprocProcessor(builders, num_workers)
        self.dependency_graph = self._get_builder_dependency_graph()
        self.has_run = []  # for bookkeeping builder runs

    # TODO: make it efficient, O(N^2) complexity at the moment,
    # might be ok(not many builders)? - KM
    def _get_builder_dependency_graph(self):
        """
        Does the following:
        1.) use targets and sources of builders to determine interdependencies
        2.) order builders according to interdependencies

        Returns:
            dict
        """
        # key = index of the builder in the self.builders list
        # value = list of indices of builders that the key depends on i.e these must run before
        # the builder corresponding to the key.
        links_dict = defaultdict(list)
        for i, bi in enumerate(self.builders):
            for j, bj in enumerate(self.builders):
                if i != j:
                    for s in bi.sources:
                        if s in bj.targets:
                            links_dict[i].append(j)
        return links_dict

    def run(self):
        """
        Does the following:
            - traverse through the builder dependency graph and does the following to
              each builder
                - connect to sources
                - get items and feed it to the processing pipeline
                - process each item
                    - supported options: serial, MPI or the builtin multiprocessing
                - collect all processed items
                - connect to the targets
                - update targets
                - finalize aka cleanup(close all connections etc)
        """
        for i in range(len(self.builders)):
            self._build_dependencies(i)

    def _build_dependencies(self, builder_id):
        """
        Run the builders by recursively traversing through the dependency graph.

        Args:
            builder_id (int): builder index
        """
        if builder_id in self.has_run:
            return
        else:
            if self.dependency_graph[builder_id]:
                for j in self.dependency_graph[builder_id]:
                    self._build_dependencies(j)
            self._run_builder(builder_id)
            self.has_run.append(builder_id)

    def _run_builder(self, builder_id):
        """
        Run builder: self.builders[builder_id]

        Args:
            builder_id (int): builder index

        Returns:

        """
        self.logger.debug("Building: {}".format(builder_id))
        self.processor.process(builder_id)



# coding: utf-8
"""
Base Builder class to define how builders need to be defined
"""
from abc import ABCMeta, abstractmethod
import logging

from monty.json import MSONable, MontyDecoder


class Builder(MSONable, metaclass=ABCMeta):
    """
    Base Builder class
    At minimum this class should implement:
    get_items - Get items from the sources
    update_targets - Updates the sources with results

    Multiprocessing and MPI processing can be used if all
    the data processing is  limited to process_items
    """

    def __init__(self, sources, targets, chunk_size=1000):
        """
        Initialize the builder the framework.

        Args:
            sources([Store]): list of source stores
            targets([Store]): list of target stores
            chunk_size(int): chunk size for processing
        """
        self.sources = sources
        self.targets = targets
        self.chunk_size = chunk_size

        self.logger = logging.getLogger(type(self).__name__)
        self.logger.addHandler(logging.NullHandler())

    def connect(self):
        """
        Connect to the builder sources and targets.
        """
        stores = self.sources + self.targets
        for s in stores:
            s.connect()

    @abstractmethod
    def get_items(self):
        """
        Returns all the items to process.

        Returns:
            generator or list of items to process
        """
        pass

    def process_item(self, item):
        """
        Process an item. Should not expect DB access as this can be run MPI
        Default behavior is to return the item.
        Args:
            item:

        Returns:
           item: an item to update
        """
        return item

    @abstractmethod
    def update_targets(self, items):
        """
        Takes a dictionary of targets and items from process item and updates them
        Can also perform other book keeping in the process such as storing gridfs oids, etc.

        Args:
            items:

        Returns:

        """
        pass

    def finalize(self, cursor=None):
        """
        Perform any final clean up.
        """
        # Close any Mongo connections.
        for store in (self.sources + self.targets):
            try:
                store.collection.database.client.close()
            except AttributeError:
                continue
        # Runner will pass iterable yielded by `self.get_items` as `cursor`. If
        # this is a Mongo cursor with `no_cursor_timeout=True` (not the
        # default), we must be explicitly kill it.
        try:
            cursor and cursor.close()
        except AttributeError:
            pass

    def __getstate__(self):
        """
        Double underscore method used by pickle to serialize this object
        This uses MSONable serialization instead
        """
        return self.as_dict()

    def __setstate__(self, d):
        """
        Double underscore method used by pickle to deserialize this object
        This uses MSONable deerialization instead
        """
        del d["@class"]
        del d["@module"]
        md = MontyDecoder()
        d = md.process_decoded(d)
        self.__init__(**d)



# coding: utf-8
"""
Tests utillities
"""
import unittest
from maggma.utils import get_mongolike, make_mongolike, put_mongolike, recursive_update


class UtilsTests(unittest.TestCase):

    def test_get_mongolike(self):
        d = {"a": [{"b": 1}, {"c": {"d": 2}}], "e": {"f": {"g": 3}}, "g": 4, "h": [5, 6]}

        self.assertEqual(get_mongolike(d, "g"), 4)
        self.assertEqual(get_mongolike(d, "e.f.g"), 3)
        self.assertEqual(get_mongolike(d, "a.0.b"), 1)
        self.assertEqual(get_mongolike(d, "a.1.c.d"), 2)
        self.assertEqual(get_mongolike(d, "h.-1"), 6)

    def test_put_mongolike(self):
        self.assertEqual(put_mongolike("e", 1), {"e": 1})
        self.assertEqual(put_mongolike("e.f.g", 1), {"e": {"f": {"g": 1}}})

    def test_make_mongolike(self):
        d = {"a": [{"b": 1}, {"c": {"d": 2}}], "e": {"f": {"g": 3}}, "g": 4, "h": [5, 6]}

        self.assertEqual(make_mongolike(d, "e.f.g", "a"), {"a": 3})
        self.assertEqual(make_mongolike(d, "e.f.g", "a.b"), {"a": {"b": 3}})
        self.assertEqual(make_mongolike(d, "a.0.b", "e.f"), {"e": {"f": 1}})

    def test_recursiveupdate(self):
        d = {"a": {"b": 3}, "c": [4]}

        recursive_update(d, {"c": [5]})
        self.assertEqual(d["c"], [5])

        recursive_update(d, {"a": {"b": 5}})
        self.assertEqual(d["a"]["b"], 5)

        recursive_update(d, {"a": {"b": [6]}})
        self.assertEqual(d["a"]["b"], [6])

        recursive_update(d, {"a": {"b": [7]}})
        self.assertEqual(d["a"]["b"], [7])



#!/usr/bin/env python
# coding utf-8

from maggma.runner import Runner
from monty.serialization import loadfn
import argparse
import logging
import sys
from maggma.utils import TqdmLoggingHandler


def main():
    parser = argparse.ArgumentParser(description="mrun is a script to run builders written using the Maggma framework.")
    parser.add_argument(
        "builder",
        help="Builder file in either json or yaml format. Can contain a list of builders or a predefined Runner")
    parser.add_argument(
        "-n",
        "--num_workers",
        type=int,
        default=0,
        help="Number of worker processes. Defaults to use as many as available.")
    parser.add_argument('-v', '--verbose', action='count', default=0, help="Controls logging level per number of v's")
    parser.add_argument(
        "--dry_run",
        action="store_true",
        default=False,
        help="Dry run loading the builder file. Does not run the builders")
    args = parser.parse_args()

    # Set Logging
    levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    level = levels[min(len(levels) - 1, args.verbose)]  # capped to number of levels
    root = logging.getLogger()
    root.setLevel(level)
    ch = TqdmLoggingHandler()
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    ch.setFormatter(formatter)
    root.addHandler(ch)

    objects = loadfn(args.builder)

    if isinstance(objects, list):
        # If this is a list of builders
        runner = Runner(objects, num_workers=args.num_workers)
    elif isinstance(objects, Runner):
        # This is a runner:
        root.info("Changing number of workers from default in input file")
        runner = Runner(objects.builders, args.num_workers)
    else:
        root.error("Couldn't properly read the builder file.")

    if not args.dry_run:
        runner.run()


if __name__ == "__main__":
    main()



# coding: utf-8
"""
More utilities for maggma. Do we need this still?
"""
import json

from pymongo import MongoClient


def get_database(cred, **mongo_client_kwargs):
    """Connect to a database given a credential dict.

    Args:
        cred (dict): {database, [host, port, username, password]}

    Returns:
        pymongo.database.Database: The database object.
    """
    # respect potential multiprocessing fork
    mc_kwargs = dict(connect=False)
    mc_kwargs.update(mongo_client_kwargs)
    conn = MongoClient(
        cred.get('host', 'localhost'),
        cred.get('port', 27017),
        **mc_kwargs)
    db = conn[cred['database']]
    if cred.get('username'):
        db.authenticate(cred['username'], cred['password'])
    return db


def get_collection(config):
    """
    Returns collection from config file

    Args:
        config(str): path to the collection config file

    Returns:
        pymongo.collection
    """
    with open(config, "r") as f:
        settings = json.load(f)
    settings["aliases_config"] = {"aliases": {}, "defaults": {}}
    db = get_database(cred=settings)
    return db[config]


class CredentialManager:

    roles = ['read', 'write', 'admin']

    def __init__(self, filepath):
        """
        Args:
            filepath (str): path to the file
        """
        with open(filepath) as f:
            self.creds = json.load(f)
            self.filepath = filepath

    def get_cred(self, spec):
        """Get DB credential dict.

        Args:
            spec (str): "<role>:<host[:port]>/<database>", where <role> is
                "read", "write", or "admin".

        Returns:
            dict: {host,port,database,username,password}

        """
        pass

    def add_cred(self, cred, role):
        """
        Add DB credential dict to `self.filepath`.

        Args:
            cred
            role
        """
        assert role in self.roles
        pass

    def ensure_cred(self, spec):
        """
        Attempt to ensure credentials as per spec.

        Generates user/pass if no existing spec match.
        Fails if host requires user/pass and cred file has neither
        an admin cred for the spec database nor a cred for the
        spec host admin db.

        Args:
            spec
        """
        pass



# coding: utf-8
"""
Tests for the Runner class
"""
import unittest
from unittest.mock import patch, MagicMock

from maggma.runner import Runner, SerialProcessor, MultiprocProcessor, MPIProcessor

__author__ = 'Shyam Dwaraknath'
__email__ = 'shyamd@lbl.gov'


class TestRunner(unittest.TestCase):
    def test_1(self):
        builder1 = MagicMock()
        builder2 = MagicMock()

        builder1.configure_mock(sources=[1, 2, 3], targets=[4])
        builder2.configure_mock(sources=[3, 4, 5], targets=[6])
        self.builders = [builder1, builder2]

        rnr = Runner(self.builders)
        self.assertEqual(rnr.dependency_graph, {1: [0]})


class TestSerialProcessor(unittest.TestCase):
    def test_process(self):

        builder = MagicMock()
        builder.configure_mock(chunk_size=10)
        builder.get_items.return_value = range(10)
        builder.process_item.side_effect = range(10, 20)

        proc = SerialProcessor([builder])

        proc.process(0)

        builder.get_items.assert_called()
        self.assertEqual(builder.process_item.call_count, 10)
        builder.update_targets.assert_called()


class TestMultiprocProcessor(unittest.TestCase):
    def setUp(self):
        builder = MagicMock()
        builder.configure_mock(chunk_size=10)
        builder.get_items.return_value = iter(range(10))
        builder.process_item.side_effect = range(10, 20)
        builder.from_dict.return_value = {}

        self.builder = builder

    def test_init(self):
        proc = MultiprocProcessor([], 3)
        self.assertEqual(proc.num_workers, 3)

    def test_setup_multithreading(self):
        
        with patch("maggma.runner.Thread") as mock_thread:
            proc = MultiprocProcessor([self.builder], num_workers=3)
            proc.builder = proc.builders[0]
            proc.setup_multithreading()
            mock_thread.assert_called()

    def test_update_targets(self):

        proc = MultiprocProcessor([self.builder], num_workers=3)

        proc.builder = self.builder
        proc.update_data_condition = MagicMock()
        proc.data = MagicMock()
        proc.update_pbar = MagicMock()
        proc.run_update_targets = MagicMock()
        proc.run_update_targets.__bool__.side_effect = [True,True,True,False]

        proc.update_targets()
        proc.run_update_targets.__bool__.assert_called()
        proc.data.clear.assert_called()
        proc.update_data_condition.wait_for.assert_called()
        proc.builder.update_targets.assert_called()

    def test_update_data_callback(self):

        proc = MultiprocProcessor([self.builder], num_workers=3)

        future = MagicMock()
        proc.data = MagicMock()
        proc.task_count = MagicMock()
        proc.update_data_condition = MagicMock()
        proc.process_pbar = MagicMock()
        proc.update_data_callback(future)

        future.result.assert_called()
        proc.update_data_condition.notify_all.assert_called()
        proc.task_count.release.assert_called()

    def test_clean_up_data(self):

        proc = MultiprocProcessor([self.builder], num_workers=3)

        proc.data = MagicMock()
        proc.update_data_condition = MagicMock()
        proc.builder = MagicMock()
        proc.update_targets_thread = MagicMock()

        proc.clean_up_data()

        proc.update_data_condition.notify_all.assert_called()
        proc.update_targets_thread.join.assert_called()

    def test_put_tasks(self):

        with patch("maggma.runner.ProcessPoolExecutor") as mock_executor:

            mock_exec_obj = mock_executor()
            proc = MultiprocProcessor([self.builder], num_workers=3)
            proc.builder = MagicMock()
            proc.task_count = MagicMock()
            cursor = [True,True,True,False]
            proc.get_pbar = cursor
            proc.put_tasks()
            proc.task_count.acquire.assert_called()


class TestMPIProcessor(unittest.TestCase):
    def setUp(self):
        builder = MagicMock()
        builder.configure_mock(chunk_size=10)
        builder.get_items.return_value = iter(range(10))
        builder.process_item.side_effect = range(10, 20)
        builder.from_dict.return_value = {}
        builder.setup_pbars([])

        self.builder = builder
        self.get_mpi_patcher = patch("maggma.runner.get_mpi")
        self.get_mpi = self.get_mpi_patcher.start()
        self.comm = MagicMock()
        self.get_mpi.return_value = self.comm, 0, 2  # comm, rank , size

    def tearDown(self):
        self.get_mpi.stop()

    def test_init(self):
        proc = MPIProcessor([self.builder])
        self.comm.barrier.assert_called()

    def test_setup_multithreading(self):

        with patch("maggma.runner.Thread") as mock_thread:
            proc = MPIProcessor([self.builder])
            proc.builder = proc.builders[0]
            proc.setup_multithreading()
            mock_thread.assert_called()

    def test_update_targets(self):

        proc = MPIProcessor([self.builder])

        proc.builder = self.builder
        proc.update_data_condition = MagicMock()
        proc.data = MagicMock()
        proc.run_update_targets = MagicMock()
        proc.run_update_targets.__bool__.side_effect = [True,True,True,False]
        proc.setup_pbars([])

        proc.update_targets()
        proc.run_update_targets.__bool__.assert_called()
        proc.data.clear.assert_called()
        proc.update_data_condition.wait_for.assert_called()
        proc.builder.update_targets.assert_called()

    def test_clean_up_data(self):

        proc = MPIProcessor([self.builder])

        proc.data = MagicMock()
        proc.update_data_condition = MagicMock()
        proc.builder = MagicMock()
        proc.update_targets_thread = MagicMock()

        proc.clean_up_data()

        proc.update_data_condition.notify_all.assert_called()
        proc.update_targets_thread.join.assert_called()

    def test_clean_up_workers(self):
        proc = MPIProcessor([self.builder])

        proc.clean_up_workers()
        self.comm.send.assert_called()
        self.assertEqual(self.comm.send.call_count, 1)

    def test_submit_item(self):

        proc = MPIProcessor([self.builder])
        proc.ranks = MagicMock()
        proc.update_data_condition = MagicMock()
        proc.data = MagicMock()
        proc.task_count = MagicMock()
        proc.setup_pbars([])
        self.comm.recv.return_value = {"type": "return", "return": "data"}

        proc.submit_item(0, {})

        self.comm.recv.assert_called()
        proc.update_data_condition.__enter__.assert_called()
        proc.data.append.assert_called()
        proc.update_data_condition.notify_all.assert_called()

        proc.ranks.append.assert_called()

    def test_put_tasks(self):
        with patch("maggma.runner.ThreadPoolExecutor") as mock_executor:
            proc = MPIProcessor([self.builder])
            proc.builder = MagicMock()
            proc.task_count = MagicMock()
            cursor = [True,True,True,False]


            proc.setup_pbars(cursor)
            proc.put_tasks(0)
            proc.task_count.acquire.assert_called()
            mock_executor.return_value.__enter__.assert_called()
            proc.task_count.acquire.assert_called()
            mock_executor.return_value.__enter__.return_value.submit.assert_called()

    def test_proccess_worker(self):
        proc = MPIProcessor([self.builder])

        self.comm.recv.side_effect = [{"type": "process", "builder_id": 0, "data": ""}, {"type": "shutdown"}]

        proc.process_worker()

        self.comm.recv.assert_called()
        self.comm.send.assert_called()
        self.builder.process_item.assert_called()


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
"""
Tests the validators
"""
import unittest
from maggma.validator import StandardValidator
from monty.json import MSONable

class ValidatorTests(unittest.TestCase):
    """
    Tests for Validators.
    """

    def test_standardvalidator(self):
        """
        Test the StandardValidator class.
        """

        class LatticeMock(MSONable):
            """
            A sample MSONable object, just for testing.
            """
            def __init__(self, a):
                self.a = a

        class SampleValidator(StandardValidator):
            """
            A sample validator, just for testing.
            """

            @property
            def schema(self):
                return {
                    "type": "object",
                    "properties":
                        {
                            "task_id": {"type": "string"},
                            "successful": {"type": "boolean"}
                        },
                    "required": ["task_id", "successful"]
                }

            @property
            def msonable_keypaths(self):
                return {"lattice": LatticeMock}

        validator = SampleValidator()

        lattice = LatticeMock(5)

        valid_doc = {
            'task_id': 'mp-test',
            'successful': True,
            'lattice': lattice.as_dict()
        }

        invalid_doc_msonable = {
            'task_id': 'mp-test',
            'successful': True,
            'lattice': ['I am not a lattice!']
        }

        invalid_doc_missing_key = {
            'task_id': 'mp-test',
            'lattice': lattice.as_dict()
        }

        invalid_doc_wrong_type = {
            'task_id': 'mp-test',
            'successful': 'true',
            'lattice': lattice.as_dict()
        }

        self.assertTrue(validator.is_valid(valid_doc))
        self.assertFalse(validator.is_valid(invalid_doc_msonable))
        self.assertFalse(validator.is_valid(invalid_doc_missing_key))
        self.assertFalse(validator.is_valid(invalid_doc_wrong_type))


# coding: utf-8
"""
Utilities to help with maggma functions
"""
import itertools
from datetime import datetime, timedelta
from pydash.utilities import to_path
from pydash.objects import set_, get, has
from pydash.objects import unset as _unset
import logging
import tqdm


class TqdmLoggingHandler(logging.Handler):
    """
    Helper to enable routing tqdm progress around logging
    """

    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


def confirm_field_index(store, fields):
    """Confirm index on store for at least one of fields

    One can't simply ensure an index exists via
    `store.collection.create_index` because a Builder must assume
    read-only access to source Stores. The MongoDB `read` built-in role
    does not include the `createIndex` action.

    Returns:
        True if an index exists for a given field
        False if not

    """
    if not isinstance(fields, list):
        fields = [fields]
    info = store.collection.index_information().values()
    for spec in (index['key'] for index in info):
        for field in fields:
            if spec[0][0] == field:
                return True
    return False


def dt_to_isoformat_ceil_ms(dt):
    """Helper to account for Mongo storing datetimes with only ms precision."""
    return (dt + timedelta(milliseconds=1)).isoformat(timespec='milliseconds')


# This lu_key prioritizes not duplicating potentially expensive item
# processing on incremental rebuilds at the expense of potentially missing a
# source document updated within 1 ms of a builder get_items call. Ensure
# appropriate builder validation.
LU_KEY_ISOFORMAT = (lambda s: datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%f"), dt_to_isoformat_ceil_ms)


def get_mongolike(d, key):
    """
    Grab a dict value using dot-notation like "a.b.c" from dict {"a":{"b":{"c": 3}}}
    Args:
        d (dict): the dictionary to search
        key (str): the key we want to grab with dot notation, e.g., "a.b.c"

    Returns:
        value from desired dict (whatever is stored at the desired key)

    """
    lead_key = key.split(".", 1)[0]
    try:
        lead_key = int(lead_key)  # for searching array data
    except:
        pass

    if "." in key:
        remainder = key.split(".", 1)[1]
        return get_mongolike(d[lead_key], remainder)
    return d[lead_key]


def put_mongolike(key, value):
    """
    Builds a dictionary with a value using mongo dot-notation

    Args:
        key (str): the key to put into using mongo notation, doesn't support arrays
        value: object
    """
    lead_key = key.split(".", 1)[0]

    if "." in key:
        remainder = key.split(".", 1)[1]
        return {lead_key: put_mongolike(remainder, value)}
    return {lead_key: value}


def make_mongolike(d, get_key, put_key):
    """
    Builds a dictionary with a value from another dictionary using mongo dot-notation

    Args:
        d (dict)L the dictionary to search
        get_key (str): the key to grab using mongo notation
        put_key (str): the key to put into using mongo notation, doesn't support arrays
    """
    return put_mongolike(put_key, get_mongolike(d, get_key))


def recursive_update(d, u):
    """
    Recursive updates d with values from u

    Args:
        d (dict): dict to update
        u (dict): updates to propogate
    """

    for k, v in u.items():
        if k in d:
            if isinstance(v, dict) and isinstance(d[k], dict):
                recursive_update(d[k], v)
            else:
                d[k] = v
        else:
            d[k] = v


def grouper(iterable, n, fillvalue=None):
    """
    Collect data into fixed-length chunks or blocks.
    """
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


def get_mpi():
    """
    Helper that returns the mpi communicator, rank and size.
    """
    try:
        from mpi4py import MPI

        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        size = comm.Get_size()
    except:
        comm = None
        rank = -1
        size = 0

    return comm, rank, size


def lazy_substitute(d, aliases):
    """
    Simple top level substitute that doesn't dive into mongo like strings
    """
    for alias, key in aliases.items():
        if key in d:
            d[alias] = d[key]
            del d[key]


def substitute(d, aliases):
    """
    Substitutes keys in dictionary
    Accepts multilevel mongo like keys
    """
    for alias, key in aliases.items():
        if has(d, key):
            set_(d, alias, get(d, key))
            unset(d, key)


def unset(d, key):
    """
    Unsets a key
    """
    _unset(d, key)
    path = to_path(key)
    for i in reversed(range(1, len(path))):
        if len(get(d, path[:i])) == 0:
            unset(d, path[:i])



# coding: utf-8
"""
Tests for advanced stores
"""
import os
import unittest
from unittest.mock import patch, MagicMock
import mongomock.collection

from maggma.stores import MemoryStore, MongoStore
from maggma.advanced_stores import *
import zlib

module_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))


class TestVaultStore(unittest.TestCase):
    """
    Test VaultStore class
    """

    def _create_vault_store(self):
        with patch('hvac.Client') as mock:

            instance = mock.return_value
            instance.auth_github.return_value = True
            instance.is_authenticated.return_value = True
            instance.read.return_value = {
                'wrap_info': None,
                'request_id': '2c72c063-2452-d1cd-19a2-91163c7395f7',
                'data': {'value': '{"db": "mg_core_prod", "host": "matgen2.lbl.gov", "username": "test", "password": "pass"}'},
                'auth': None,
                'warnings': None,
                'renewable': False,
                'lease_duration': 2764800, 'lease_id': ''
            }
            v = VaultStore("test_coll", "secret/matgen/maggma")

        return v

    def test_vault_init(self):
        """
        Test initing a vault store using a mock hvac client
        """
        os.environ['VAULT_ADDR'] = "https://fake:8200/"
        os.environ['VAULT_TOKEN'] = "dummy"

        v = self._create_vault_store()
        # Just test that we successfully instantiated
        assert isinstance(v, MongoStore)

    def test_vault_github_token(self):
        """
        Test using VaultStore with GITHUB_TOKEN and mock hvac
        """
        # Save token in env
        os.environ['VAULT_ADDR'] = "https://fake:8200/"
        os.environ['GITHUB_TOKEN'] = "dummy"

        v = self._create_vault_store()
        # Just test that we successfully instantiated
        assert isinstance(v, MongoStore)

    def test_vault_missing_env(self):
        """
        Test VaultStore should raise an error if environment is not set
        """
        del os.environ['VAULT_TOKEN']
        del os.environ['VAULT_ADDR']
        del os.environ['GITHUB_TOKEN']

        # Create should raise an error
        with self.assertRaises(RuntimeError):
            self._create_vault_store()


class TestS3Store(unittest.TestCase):

    def setUp(self):
        self.index = MemoryStore("index'")
        with patch("boto3.resource") as mock_resource:
            mock_resource.return_value = MagicMock()
            mock_resource("s3").list_buckets.return_value = ["bucket1", "bucket2"]
            self.s3store = AmazonS3Store(self.index, "bucket1")
            self.s3store.connect()

    def test_qeuery_one(self):
        self.s3store.s3_bucket.Object.return_value = MagicMock()
        self.s3store.s3_bucket.Object().get.return_value = '{"task_id": "mp-1", "data": "asd"}'
        self.index.update([{"task_id": "mp-1"}])
        self.assertEqual(self.s3store.query_one(criteria={"task_id": "mp-2"}), None)
        self.assertEqual(self.s3store.query_one(criteria={"task_id": "mp-1"})["data"], "asd")

        self.s3store.s3_bucket.Object().get.return_value = zlib.compress('{"task_id": "mp-3", "data": "sdf"}'.encode())
        self.index.update([{"task_id": "mp-3", "compression": "zlib"}])
        self.assertEqual(self.s3store.query_one(criteria={"task_id": "mp-3"})["data"], "sdf")

    def test_update(self):

        self.s3store.update([{"task_id": "mp-1", "data": "asd"}])
        self.assertEqual(self.s3store.s3_bucket.put_object.call_count, 1)
        called_kwargs = self.s3store.s3_bucket.put_object.call_args[1]
        self.assertEqual(self.s3store.s3_bucket.put_object.call_count, 1)
        self.assertEqual(called_kwargs["Key"], "mp-1")
        self.assertTrue(len(called_kwargs["Body"]) > 0)
        self.assertEqual(called_kwargs["Metadata"]["task_id"], "mp-1")

    def test_update_compression(self):
        self.s3store.update([{"task_id": "mp-1", "data": "asd"}], compress=True)
        self.assertEqual(self.s3store.s3_bucket.put_object.call_count, 1)
        called_kwargs = self.s3store.s3_bucket.put_object.call_args[1]
        self.assertEqual(self.s3store.s3_bucket.put_object.call_count, 1)
        self.assertEqual(called_kwargs["Key"], "mp-1")
        self.assertTrue(len(called_kwargs["Body"]) > 0)
        self.assertEqual(called_kwargs["Metadata"]["task_id"], "mp-1")
        self.assertEqual(called_kwargs["Metadata"]["compression"], "zlib")


class TestAliasingStore(unittest.TestCase):

    def setUp(self):
        self.memorystore = MemoryStore("test")
        self.memorystore.connect()
        self.aliasingstore = AliasingStore(
            self.memorystore, {"a": "b", "c.d": "e", "f": "g.h"})

    def test_query(self):

        d = [{"b": 1}, {"e": 2}, {"g": {"h": 3}}]
        self.memorystore.collection.insert_many(d)

        self.assertTrue("a" in list(self.aliasingstore.query(
            criteria={"a": {"$exists": 1}}))[0])
        self.assertTrue("c" in list(self.aliasingstore.query(
            criteria={"c.d": {"$exists": 1}}))[0])
        self.assertTrue("d" in list(self.aliasingstore.query(
            criteria={"c.d": {"$exists": 1}}))[0].get("c", {}))
        self.assertTrue("f" in list(self.aliasingstore.query(
            criteria={"f": {"$exists": 1}}))[0])

    def test_update(self):

        self.aliasingstore.update([{"task_id": "mp-3", "a": 4}, {"task_id": "mp-4",
                                                                 "c": {"d": 5}}, {"task_id": "mp-5", "f": 6}])
        self.assertEqual(list(self.aliasingstore.query(criteria={"task_id": "mp-3"}))[0]["a"], 4)
        self.assertEqual(list(self.aliasingstore.query(criteria={"task_id": "mp-4"}))[0]["c"]["d"], 5)
        self.assertEqual(list(self.aliasingstore.query(criteria={"task_id": "mp-5"}))[0]["f"], 6)

        self.assertEqual(list(self.aliasingstore.store.query(criteria={"task_id": "mp-3"}))[0]["b"], 4)
        self.assertEqual(list(self.aliasingstore.store.query(criteria={"task_id": "mp-4"}))[0]["e"], 5)
        self.assertEqual(list(self.aliasingstore.store.query(criteria={"task_id": "mp-5"}))[0]["g"]["h"], 6)

    def test_substitute(self):
        aliases = {"a": "b", "c.d": "e", "f": "g.h"}

        d = {"b": 1}
        substitute(d, aliases)
        self.assertTrue("a" in d)

        d = {"e": 1}
        substitute(d, aliases)
        self.assertTrue("c" in d)
        self.assertTrue("d" in d.get("c", {}))

        d = {"g": {"h": 4}}
        substitute(d, aliases)
        self.assertTrue("f" in d)

        d = None
        substitute(d, aliases)
        self.assertTrue(d is None)


class TestSandboxStore(unittest.TestCase):

    def setUp(self):
        self.store = MemoryStore()
        self.sandboxstore = SandboxStore(self.store, sandbox="test")

    def test_connect(self):
        self.assertEqual(self.sandboxstore.collection, None)
        self.sandboxstore.connect()
        self.assertIsInstance(self.sandboxstore.collection, mongomock.collection.Collection)

    def test_query(self):
        self.sandboxstore.connect()
        self.sandboxstore.collection.insert_one({"a": 1, "b": 2, "c": 3})
        self.assertEqual(self.sandboxstore.query_one(properties=["a"])['a'], 1)

        self.sandboxstore.collection.insert_one({"a": 2, "b": 2, "sbxn": ["test"]})
        self.assertEqual(self.sandboxstore.query_one(properties=["b"],
                                                     criteria={"a": 2})['b'], 2)

        self.sandboxstore.collection.insert_one({"a": 3, "b": 2, "sbxn": ["not_test"]})
        self.assertEqual(self.sandboxstore.query_one(properties=["c"],
                                                   criteria={"a": 3}), None)

    def test_distinct(self):
        self.sandboxstore.connect()
        self.sandboxstore.collection.insert_one({"a": 1, "b": 2, "c": 3})
        self.assertEqual(self.sandboxstore.distinct("a"), [1])

        self.sandboxstore.collection.insert_one({"a": 4, "d": 5, "e": 6, "sbxn": ["test"]})
        self.assertEqual(self.sandboxstore.distinct("a"), [1, 4])

        self.sandboxstore.collection.insert_one({"a": 7, "d": 8, "e": 9, "sbxn": ["not_test"]})
        self.assertEqual(self.sandboxstore.distinct("a"), [1, 4])

    def test_update(self):
        self.sandboxstore.connect()
        self.sandboxstore.update([{"e": 6, "d": 4}], key="e")
        self.assertEqual(self.sandboxstore.query(criteria={"d": {"$exists": 1}}, properties=["d"])[0]["d"], 4)
        self.assertEqual(self.sandboxstore.collection.find_one({"e": 6})["sbxn"], ["test"])
        self.sandboxstore.update([{"e": 7, "sbxn": ["core"]}], key="e")
        self.assertEqual(set(self.sandboxstore.query_one(
            criteria={"e": 7})["sbxn"]), {"test", "core"})

    def tearDown(self):
        if self.sandboxstore.collection:
            self.sandboxstore.collection.drop()


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
"""
Tests for the base Stores
"""
import os
import glob
import unittest
import numpy as np
import mongomock.collection
import pymongo.collection
import numpy.testing.utils as nptu
from maggma.stores import *

module_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
db_dir = os.path.abspath(os.path.join(module_dir, "..", "..", "test_files", "settings_files"))
test_dir = os.path.abspath(os.path.join(module_dir, "..", "..", "test_files", "test_set"))


class TestMongoStore(unittest.TestCase):

    def setUp(self):
        self.mongostore = MongoStore("maggma_test", "test")
        self.mongostore.connect()

    def test_connect(self):
        mongostore = MongoStore("maggma_test", "test")
        self.assertEqual(mongostore.collection, None)
        mongostore.connect()
        self.assertIsInstance(mongostore.collection, pymongo.collection.Collection)

    def test_query(self):
        self.mongostore.collection.insert({"a": 1, "b": 2, "c": 3})
        self.assertEqual(self.mongostore.query_one(properties=["a"])["a"], 1)
        self.assertEqual(self.mongostore.query_one(properties=["a"])['a'], 1)
        self.assertEqual(self.mongostore.query_one(properties=["b"])['b'], 2)
        self.assertEqual(self.mongostore.query_one(properties=["c"])['c'], 3)

    def test_distinct(self):
        self.mongostore.collection.insert({"a": 1, "b": 2, "c": 3})
        self.mongostore.collection.insert({"a": 4, "d": 5, "e": 6})
        self.assertEqual(self.mongostore.distinct("a"), [1, 4])

        # Test list distinct functionality
        self.mongostore.collection.insert({"a": 4, "d": 6, "e": 7})
        self.mongostore.collection.insert({"a": 4, "d": 6})
        ad_distinct = self.mongostore.distinct(["a", "d"])
        self.assertTrue(len(ad_distinct), 3)
        self.assertTrue({"a": 4, "d": 6} in ad_distinct)
        self.assertTrue({"a": 1} in ad_distinct)
        self.assertEqual(len(self.mongostore.distinct(["d", "e"], {"a": 4})), 3)
        all_exist = self.mongostore.distinct(["a", "b"], all_exist=True)
        self.assertEqual(len(all_exist), 1)
        all_exist2 = self.mongostore.distinct(["a", "e"], all_exist=True, criteria={"d": 6})
        self.assertEqual(len(all_exist2), 1)

    def test_update(self):
        self.mongostore.update([{"e": 6, "d": 4}], key="e")
        self.assertEqual(self.mongostore.query(criteria={"d": {"$exists": 1}}, properties=["d"])[0]["d"], 4)

        self.mongostore.update([{"e": 7, "d": 8, "f": 9}], key=["d", "f"])
        self.assertEqual(self.mongostore.query_one(criteria={"d": 8, "f": 9}, properties=["e"])["e"], 7)
        self.mongostore.update([{"e": 11, "d": 8, "f": 9}], key=["d", "f"])
        self.assertEqual(self.mongostore.query_one(criteria={"d": 8, "f": 9}, properties=["e"])["e"], 11)

    def test_groupby(self):
        self.mongostore.collection.drop()
        self.mongostore.update(
            [{
                "e": 7,
                "d": 9,
                "f": 9
            }, {
                "e": 7,
                "d": 9,
                "f": 10
            }, {
                "e": 8,
                "d": 9,
                "f": 11
            }, {
                "e": 9,
                "d": 10,
                "f": 12
            }],
            key="f")
        data = list(self.mongostore.groupby("d"))
        self.assertEqual(len(data), 2)
        grouped_by_9 = [g['docs'] for g in data if g['_id']['d'] == 9][0]
        self.assertEqual(len(grouped_by_9), 3)
        grouped_by_10 = [g['docs'] for g in data if g['_id']['d'] == 10][0]
        self.assertEqual(len(grouped_by_10), 1)

        data = list(self.mongostore.groupby(["e", "d"]))
        self.assertEqual(len(data), 3)

    def test_from_db_file(self):
        ms = MongoStore.from_db_file(os.path.join(db_dir, "db.json"))
        self.assertEqual(ms.collection_name, "tmp")

    def test_from_collection(self):
        ms = MongoStore.from_db_file(os.path.join(db_dir, "db.json"))
        ms.connect()

        other_ms = MongoStore.from_collection(ms._collection)
        self.assertEqual(ms.collection_name, other_ms.collection_name)
        self.assertEqual(ms.database, other_ms.database)

    def tearDown(self):
        if self.mongostore.collection:
            self.mongostore.collection.drop()


class TestMemoryStore(unittest.TestCase):

    def setUp(self):
        self.memstore = MemoryStore()

    def test(self):
        self.assertEqual(self.memstore.collection, None)
        self.memstore.connect()
        self.assertIsInstance(self.memstore.collection, mongomock.collection.Collection)

    def test_groupby(self):
        self.memstore.connect()
        self.memstore.update(
            [{
                "e": 7,
                "d": 9,
                "f": 9
            }, {
                "e": 7,
                "d": 9,
                "f": 10
            }, {
                "e": 8,
                "d": 9,
                "f": 11
            }, {
                "e": 9,
                "d": 10,
                "f": 12
            }],
            key="f")
        data = list(self.memstore.groupby("d"))
        self.assertEqual(len(data), 2)
        grouped_by_9 = [g['docs'] for g in data if g['_id']['d'] == 9][0]
        self.assertEqual(len(grouped_by_9), 3)
        grouped_by_10 = [g['docs'] for g in data if g['_id']['d'] == 10][0]
        self.assertEqual(len(grouped_by_10), 1)

        data = list(self.memstore.groupby(["e", "d"]))
        self.assertEqual(len(data), 3)


class TestJsonStore(unittest.TestCase):

    def test(self):
        files = []
        for f in ["a.json", "b.json"]:
            files.append(os.path.join(test_dir, f))

        jsonstore = JSONStore(files)
        jsonstore.connect()
        self.assertEqual(len(list(jsonstore.query())), 20)

        jsonstore = JSONStore(os.path.join(test_dir, "c.json.gz"))
        jsonstore.connect()
        self.assertEqual(len(list(jsonstore.query())), 20)


class TestGridFSStore(unittest.TestCase):

    def setUp(self):
        self.gStore = GridFSStore("maggma_test", "test", key="task_id")
        self.gStore.connect()

    def test_update(self):
        data1 = np.random.rand(256)
        data2 = np.random.rand(256)
        # Test metadata storage
        self.gStore.update([{"task_id": "mp-1", "data": data1}])
        self.assertTrue(self.gStore._files_collection.find_one({"metadata.task_id": "mp-1"}))

        # Test storing data
        self.gStore.update([{"task_id": "mp-1", "data": data2}])
        self.assertEqual(len(list(self.gStore.query({"task_id": "mp-1"}))), 1)
        self.assertTrue("task_id" in self.gStore.query_one({"task_id": "mp-1"}))
        nptu.assert_almost_equal(self.gStore.query_one({"task_id": "mp-1"})["data"], data2, 7)

        # Test storing compressed data
        self.gStore = GridFSStore("maggma_test", "test", key="task_id", compression=True)
        self.gStore.connect()
        self.gStore.update([{"task_id": "mp-1", "data": data1}])
        self.assertTrue(self.gStore._files_collection.find_one({"metadata.compression": "zlib"}))
        nptu.assert_almost_equal(self.gStore.query_one({"task_id": "mp-1"})["data"], data1, 7)

    def test_query(self):
        data1 = np.random.rand(256)
        data2 = np.random.rand(256)
        tic = datetime(2018, 4, 12, 16)
        self.gStore.update([{"task_id": "mp-1", "data": data1}])
        self.gStore.update([{"task_id": "mp-2", "data": data2, self.gStore.lu_field: tic}], update_lu=False)

        doc = self.gStore.query_one(criteria={"task_id": "mp-1"})
        nptu.assert_almost_equal(doc["data"], data1, 7)

        doc = self.gStore.query_one(criteria={"task_id": "mp-2"})
        nptu.assert_almost_equal(doc["data"], data2, 7)
        self.assertTrue(self.gStore.lu_field in doc)

        self.assertEqual(self.gStore.query_one(criteria={"task_id": "mp-3"}), None)

    @unittest.skip
    def test_distinct(self):
        # TODO
        pass

    def tearDown(self):
        if self.gStore.collection:
            self.gStore._files_collection.drop()
            self.gStore._chunks_collection.drop()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import os

from setuptools import setup, find_packages

module_dir = os.path.dirname(os.path.abspath(__file__))

if __name__ == "__main__":
    setup(
        name='maggma',
        version='0.6.0',
        description='MongoDB aggregation machine',
        long_description=open(os.path.join(module_dir, 'README.md')).read(),
        url='https://github.com/materialsproject/maggma',
        author='MP Team',
        author_email='matproj-develop@googlegroups.com',
        license='modified BSD',
        packages=find_packages(),
        package_data={},
        zip_safe=False,
        install_requires=['pymongo>=3.4.0', 'mongomock>=3.8.0', 'monty>=0.9.8',
                          'smoqe==0.1.3', 'PyYAML==3.12', 'pydash==4.1.0', 'tqdm>=4.19.6'],
        extras_require={"mpi": ["mpi4py>=2.0.0"]},
        classifiers=["Programming Language :: Python :: 3",
                     "Programming Language :: Python :: 3.6",
                     'Development Status :: 2 - Pre-Alpha',
                     'Intended Audience :: Science/Research',
                     'Intended Audience :: System Administrators',
                     'Intended Audience :: Information Technology',
                     'Operating System :: OS Independent',
                     'Topic :: Other/Nonlisted Topic',
                     'Topic :: Database :: Front-Ends',
                     'Topic :: Scientific/Engineering'],


        entry_points={
            'console_scripts': [
                'mrun = maggma.cli.mrun:main'
                ]
            },
        test_suite='nose.collector',
        tests_require=['nose']
    )



# coding: utf-8
"""
Module containing various definitions of Stores.
Stores are a default access pattern to data and provide
various utillities
"""
from abc import ABCMeta, abstractmethod
import copy
from datetime import datetime
import json
import zlib

import mongomock
import pymongo
import gridfs
from itertools import groupby
from operator import itemgetter
from pymongo import MongoClient, DESCENDING
from pydash import identity

from monty.json import MSONable, jsanitize, MontyDecoder
from monty.io import zopen
from monty.serialization import loadfn
from maggma.utils import LU_KEY_ISOFORMAT, confirm_field_index


class Store(MSONable, metaclass=ABCMeta):
    """
    Abstract class for a data Store
    Defines the interface for all data going in and out of a Builder
    """

    def __init__(self, key="task_id", lu_field='last_updated', lu_type="datetime", validator=None):
        """
        Args:
            key (str): master key to index on
            lu_field (str): 'last updated' field name
            lu_type (tuple): the date/time format for the lu_field. Can be "datetime" or "isoformat"
        """
        self.key = key
        self.lu_field = lu_field
        self.lu_type = lu_type
        self.lu_func = LU_KEY_ISOFORMAT if lu_type == "isoformat" else (identity, identity)
        self.validator = validator

    @property
    @abstractmethod
    def collection(self):
        """
        Returns a handle to the pymongo collection object
        Not guaranteed to exist in the future
        """
        pass

    @abstractmethod
    def connect(self, force_reset=False):
        """
        Connect to the source data
        """
        pass

    @abstractmethod
    def close(self):
        """
        Closes any connections
        """
        pass

    @abstractmethod
    def query(self, properties=None, criteria=None, **kwargs):
        """
        Queries the Store for a set of properties
        """
        pass

    @abstractmethod
    def query_one(self, properties=None, criteria=None, **kwargs):
        """
        Get one property from the store
        """
        pass

    @abstractmethod
    def distinct(self, key, criteria=None, **kwargs):
        """
        Get all distinct values for a key
        """
        pass

    @abstractmethod
    def update(self, docs, update_lu=True, key=None, **kwargs):
        """
        Update docs into the store
        """
        pass

    @abstractmethod
    def ensure_index(self, key, unique=False, **kwargs):
        """
        Tries to create and index
        Args:
            key (string): single key to index
            unique (bool): Whether or not this index contains only unique keys

        Returns:
            bool indicating if the index exists/was created
        """
        pass

    @abstractmethod
    def groupby(self, keys, properties=None, criteria=None, **kwargs):
        """
        Simple grouping function that will group documents
        by keys.

        Args:
            keys (list or string): fields to group documents
            properties (list): properties to return in grouped documents
            criteria (dict): filter for documents to group

        Returns:
            command cursor corresponding to grouped documents

            elements of the command cursor have the structure:
            {'_id': {"KEY_1": value_1, "KEY_2": value_2 ...,
             'docs': [list_of_documents corresponding to key values]}

        """
        pass

    @property
    def last_updated(self):
        doc = next(self.query(properties=[self.lu_field]).sort([(self.lu_field, pymongo.DESCENDING)]).limit(1), None)
        # Handle when collection has docs but `NoneType` lu_field.
        return (self.lu_func[0](doc[self.lu_field]) if (doc and doc[self.lu_field]) else datetime.min)

    def lu_filter(self, targets):
        """Creates a MongoDB filter for new documents.

        By "new", we mean documents in this Store that were last updated later
        than any document in targets.

        Args:
            targets (list): A list of Stores

        """
        if isinstance(targets, Store):
            targets = [targets]

        lu_list = [t.last_updated for t in targets]
        return {self.lu_field: {"$gt": self.lu_func[1](max(lu_list))}}

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.lu_field, ))

    def __getstate__(self):
        return self.as_dict()

    def __setstate__(self, d):
        del d["@class"]
        del d["@module"]
        md = MontyDecoder()
        d = md.process_decoded(d)
        self.__init__(**d)


class Mongolike(object):
    """
    Mixin class that allows for basic mongo functionality
    """

    @property
    def collection(self):
        return self._collection

    def query(self, properties=None, criteria=None, **kwargs):
        """
        Function that gets data from MongoStore with property focus.

        Args:
            properties (list or dict): list of properties to return
                or dictionary with {"property": 1} type structure
                from standard mongo Collection.find syntax
            criteria (dict): filter for query, matches documents
                against key-value pairs
            **kwargs (kwargs): further kwargs to Collection.find
        """
        if isinstance(properties, list):
            properties = {p: 1 for p in properties}
        return self.collection.find(filter=criteria, projection=properties, **kwargs)

    def query_one(self, properties=None, criteria=None, **kwargs):
        """
        Function that gets a single from MongoStore with property focus.
        Returns None if nothing matches

        Args:
            properties (list or dict): list of properties to return
                or dictionary with {"property": 1} type structure
                from standard mongo Collection.find syntax
            criteria (dict): filter for query, matches documents
                against key-value pairs
            **kwargs (kwargs): further kwargs to Collection.find_one
        """
        if isinstance(properties, list):
            properties = {p: 1 for p in properties}
        return self.collection.find_one(filter=criteria, projection=properties, **kwargs)

    def ensure_index(self, key, unique=False, **kwargs):
        """
        Wrapper for pymongo.Collection.ensure_index
        """
        if "background" not in kwargs:
            kwargs["background"] = True

        if confirm_field_index(self.collection,key):
            return True
        else:
            return self.collection.create_index(key, unique=unique, **kwargs)

    def update(self, docs, update_lu=True, key=None, **kwargs):
        """
        Function to update associated MongoStore collection.

        Args:
            docs: list of documents
        """

        bulk = self.collection.initialize_ordered_bulk_op()

        for d in docs:

            d = jsanitize(d, allow_bson=True)

            # document-level validation is optional
            validates = True
            if self.validator:
                validates = self.validator.is_valid(d)
                if not validates:
                    if self.validator.strict:
                        raise ValueError('Document failed to validate: {}'.format(d))
                    else:
                        self.logger.error('Document failed to validate: {}'.format(d))

            if validates:
                if isinstance(key, list):
                    search_doc = {k: d[k] for k in key}
                elif key:
                    search_doc = {key: d[key]}
                else:
                    search_doc = {self.key: d[self.key]}
                if update_lu:
                    d[self.lu_field] = datetime.utcnow()
                bulk.find(search_doc).upsert().replace_one(d)

        bulk.execute()

    def distinct(self, key, criteria=None, all_exist=False, **kwargs):
        """
        Function get to get all distinct values of a certain key in
        a mongolike store.  May take a single key or a list of keys

        Args:
            key (mongolike key or list of mongolike keys): key or keys
                for which to find distinct values or sets of values.
            criteria (filter criteria): criteria for filter
            all_exist (bool): whether to ensure all keys in list exist
                in each document, defaults to False
            **kwargs (kwargs): kwargs corresponding to collection.distinct
        """
        if isinstance(key, list):
            criteria = criteria if criteria else {}
            # Update to ensure keys are there
            if all_exist:
                criteria.update({k: {"$exists": True} for k in key if k not in criteria})

            results = []
            for d in self.groupby(key, properties=key, criteria=criteria):
                results.append(d["_id"])
            return results

        else:
            return self.collection.distinct(key, filter=criteria, **kwargs)

    def close(self):
        self.collection.database.client.close()


class MongoStore(Mongolike, Store):
    """
    A Store that connects to a Mongo collection
    """

    def __init__(self, database, collection_name, host="localhost", port=27017, username="", password="", **kwargs):
        """
        Args:
            database (str): database name
            collection (str): collection name
            host (str): hostname for mongo db
            port (int): tcp port for mongo db
            username (str): username for mongo db
            password (str): password for mongo db
        """
        self.database = database
        self.collection_name = collection_name
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self._collection = None
        self.kwargs = kwargs
        super(MongoStore, self).__init__(**kwargs)

    def connect(self, force_reset=False):
        if not self._collection or force_reset:
            conn = MongoClient(self.host, self.port)
            db = conn[self.database]
            if self.username is not "":
                db.authenticate(self.username, self.password)
            self._collection = db[self.collection_name]

    def __hash__(self):
        return hash((self.database, self.collection_name, self.lu_field))

    @classmethod
    def from_db_file(cls, filename):
        """
        Convenience method to construct MongoStore from db_file
        """
        kwargs = loadfn(filename)
        if "collection" in kwargs:
            kwargs["collection_name"] = kwargs.pop("collection")
        # Get rid of aliases from traditional query engine db docs
        kwargs.pop("aliases", None)
        return cls(**kwargs)

    def groupby(self, keys, properties=None, criteria=None, allow_disk_use=True, **kwargs):
        """
        Simple grouping function that will group documents
        by keys.

        Args:
            keys (list or string): fields to group documents
            properties (list): properties to return in grouped documents
            criteria (dict): filter for documents to group
            allow_disk_use (bool): whether to allow disk use in aggregation

        Returns:
            command cursor corresponding to grouped documents

            elements of the command cursor have the structure:
            {'_id': {"KEY_1": value_1, "KEY_2": value_2 ...,
             'docs': [list_of_documents corresponding to key values]}

        """
        pipeline = []
        if criteria is not None:
            pipeline.append({"$match": criteria})

        if properties is not None:
            pipeline.append({"$project": {p: 1 for p in properties}})

        if isinstance(keys, str):
            keys = [keys]

        group_id = {key: "${}".format(key) for key in keys}
        pipeline.append({"$group": {"_id": group_id, "docs": {"$push": "$$ROOT"}}})

        return self.collection.aggregate(pipeline, allowDiskUse=allow_disk_use)

    @classmethod
    def from_collection(cls, collection, **kwargs):
        """
        Generates a MongoStore from a pymongo collection object
        This is not a fully safe operation as it gives dummy information to the MongoStore
        As a result, this will not serialize and can not reset its connection
        """
        # TODO: How do we make this safer?
        coll_name = collection.name
        db_name = collection.database.name

        store = cls(db_name, coll_name, **kwargs)
        store._collection = collection
        return store


class MemoryStore(Mongolike, Store):
    """
    An in-memory Store that functions similarly
    to a MongoStore
    """

    def __init__(self, name="memory_db", **kwargs):
        self.name = name
        self._collection = None
        self.kwargs = kwargs
        super(MemoryStore, self).__init__(**kwargs)

    def connect(self, force_reset=False):
        if not self._collection or force_reset:
            self._collection = mongomock.MongoClient().db[self.name]

    def __hash__(self):
        return hash((self.name, self.lu_field))

    def groupby(self, keys, properties=None, criteria=None, **kwargs):
        """
        Simple grouping function that will group documents
        by keys.

        Args:
            keys (list or string): fields to group documents
            properties (list): properties to return in grouped documents
            criteria (dict): filter for documents to group
            allow_disk_use (bool): whether to allow disk use in aggregation

        Returns:
            command cursor corresponding to grouped documents

            elements of the command cursor have the structure:
            {'_id': {"KEY_1": value_1, "KEY_2": value_2 ...,
             'docs': [list_of_documents corresponding to key values]}

        """
        keys = keys if isinstance(keys, list) else [keys]

        input_data = list(self.query(properties=keys, criteria=criteria))

        if len(keys) > 1:
            grouper = itemgetter(*keys)
            for key, grp in groupby(sorted(input_data, key=grouper), grouper):
                temp_dict = {"_id": zip(keys, key), "docs": list(grp)}
                yield temp_dict
        else:
            grouper = itemgetter(*keys)
            for key, grp in groupby(sorted(input_data, key=grouper), grouper):
                temp_dict = {"_id": {keys[0]: key}, "docs": list(grp)}
                yield temp_dict

    def update(self, docs, update_lu=True, key=None, **kwargs):
        """
        Function to update associated MongoStore collection.

        Args:
            docs: list of documents
        """

        for d in docs:

            d = jsanitize(d, allow_bson=True)

            # document-level validation is optional
            validates = True
            if self.validator:
                validates = self.validator.is_valid(d)
                if not validates:
                    if self.validator.strict:
                        raise ValueError('Document failed to validate: {}'.format(d))
                    else:
                        self.logger.error('Document failed to validate: {}'.format(d))

            if validates:
                if isinstance(key, list):
                    search_doc = {k: d[k] for k in key}
                elif key:
                    search_doc = {key: d[key]}
                else:
                    search_doc = {self.key: d[self.key]}
                if update_lu:
                    d[self.lu_field] = datetime.utcnow()
                self.collection.insert_one(d)


class JSONStore(MemoryStore):
    """
    A Store for access to a single or multiple JSON files
    """

    def __init__(self, paths, **kwargs):
        """
        Args:
            paths (str or list): paths for json files to
                turn into a Store
        """
        paths = paths if isinstance(paths, (list, tuple)) else [paths]
        self.paths = paths
        self.kwargs = kwargs
        super(JSONStore, self).__init__("collection", **kwargs)

    def connect(self, force_reset=False):
        super(JSONStore, self).connect(force_reset=force_reset)
        for path in self.paths:
            with zopen(path) as f:
                data = f.read()
                data = data.decode() if isinstance(data, bytes) else data
                objects = json.loads(data)
                objects = [objects] if not isinstance(objects, list) else objects
                self.collection.insert_many(objects)

    def __hash__(self):
        return hash((*self.paths, self.lu_field))


class DatetimeStore(MemoryStore):
    """Utility store intended for use with `Store.lu_filter`."""

    def __init__(self, dt, **kwargs):
        """
        Args:
            dt (Datetime): Datetime to set
        """
        self.__dt = dt
        self.kwargs = kwargs
        super(DatetimeStore, self).__init__("date", **kwargs)

    def connect(self, force_reset=False):
        super(DatetimeStore, self).connect(force_reset)
        self.collection.insert_one({self.lu_field: self.__dt})


class GridFSStore(Store):
    """
    A Store for GrdiFS backend. Provides a common access method consistent with other stores
    """

    # https://github.com/mongodb/specifications/
    #   blob/master/source/gridfs/gridfs-spec.rst#terms
    #   (Under "Files collection document")
    files_collection_fields = (
        "_id", "length", "chunkSize", "uploadDate", "md5", "filename",
        "contentType", "aliases", "metadata")

    def __init__(self, database, collection_name, host="localhost", port=27017, username="", password="", compression=False, **kwargs):

        self.database = database
        self.collection_name = collection_name
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self._collection = None
        self.compression = compression
        self.kwargs = kwargs
        self.meta_keys = set()

        if "key" not in kwargs:
            kwargs["key"] = "_oid"

        super(GridFSStore, self).__init__(**kwargs)

    def connect(self, force_reset=False):
        conn = MongoClient(self.host, self.port)
        if not self._collection or force_reset:
            db = conn[self.database]
            if self.username is not "":
                db.authenticate(self.username, self.password)

            self._collection = gridfs.GridFS(db, self.collection_name)
            self._files_collection = db["{}.files".format(self.collection_name)]
            self._chunks_collection = db["{}.chunks".format(self.collection_name)]

    @property
    def collection(self):
        # TODO: Should this return the real MongoCollection or the GridFS
        return self._collection

    @classmethod
    def transform_criteria(cls, criteria):
        """
        Allow client to not need to prepend 'metadata.' to query fields.
        Args:
            criteria (dict): Query criteria
        """
        for field in criteria:
            if (field not in cls.files_collection_fields
                    and not field.startswith('metadata.')):
                criteria['metadata.' + field] = copy.copy(criteria[field])
                del criteria[field]

    def query(self, properties=None, criteria=None, **kwargs):
        """
        Function that gets data from GridFS. This store ignores all
        property projections as its designed for whole document access

        Args:
            properties (list or dict): This will be ignored by the GridFS
                Store
            criteria (dict): filter for query, matches documents
                against key-value pairs
            **kwargs (kwargs): further kwargs to Collection.find
        """
        if isinstance(criteria, dict):
            self.transform_criteria(criteria)
        for f in self.collection.find(filter=criteria, **kwargs):
            data = f.read()

            metadata = f.metadata
            if metadata.get("compression", "") == "zlib":
                data = zlib.decompress(data).decode("UTF-8")

            try:
                data = json.loads(data)
            except:
                pass
            yield data

    def query_one(self, properties=None, criteria=None, **kwargs):
        """
        Function that gets a single document from GridFS. This store
        ignores all property projections as its designed for whole
        document access

        Args:
            properties (list or dict): This will be ignored by the GridFS
                Store
            criteria (dict): filter for query, matches documents
                against key-value pairs
            **kwargs (kwargs): further kwargs to Collection.find
        """
        return next(self.query(criteria=criteria, **kwargs), None)

    def distinct(self, key, criteria=None, all_exist=False, **kwargs):
        """
        Function get to get all distinct values of a certain key in
        a mongolike store.  May take a single key or a list of keys

        Args:
            key (mongolike key or list of mongolike keys): key or keys
                for which to find distinct values or sets of values.
            criteria (filter criteria): criteria for filter
            all_exist (bool): whether to ensure all keys in list exist
                in each document, defaults to False
            **kwargs (kwargs): kwargs corresponding to collection.distinct
        """
        if isinstance(key, list):
            criteria = criteria if criteria else {}
            # Update to ensure keys are there
            if all_exist:
                criteria.update({k: {"$exists": True} for k in key if k not in criteria})

            results = []
            for d in self.groupby(key, properties=key, criteria=criteria):
                results.append(d["_id"])
            return results

        else:
            if criteria:
                self.transform_criteria(criteria)
            # Transfor to metadata subfield if not supposed to be in gridfs main fields
            if key not in self.files_collection_fields:
                key = "metadata.{}".format(key)

            return self._files_collection.distinct(key, filter=criteria, **kwargs)

    def groupby(self, keys, properties=None, criteria=None, allow_disk_use=True, **kwargs):
        """
        Simple grouping function that will group documents
        by keys.

        Args:
            keys (list or string): fields to group documents
            properties (list): properties to return in grouped documents
            criteria (dict): filter for documents to group
            allow_disk_use (bool): whether to allow disk use in aggregation

        Returns:
            command cursor corresponding to grouped documents

            elements of the command cursor have the structure:
            {'_id': {"KEY_1": value_1, "KEY_2": value_2 ...,
             'docs': [list_of_documents corresponding to key values]}

        """
        pipeline = []
        if criteria is not None:
            self.transform_criteria(criteria)
            pipeline.append({"$match": criteria})

        if properties is not None:
            properties = [p if p in self.files_collection_fields else "metadata.{}".format(p) for p in properties]
            pipeline.append({"$project": {p: 1 for p in properties}})

        if isinstance(keys, str):
            keys = [keys]

        # ensure propper naming for keys in and outside of metadata
        keys = [k if k in self.files_collection_fields else "metadata.{}".format(k) for k in key]

        group_id = {key: "${}".format(key) for key in keys}
        pipeline.append({"$group": {"_id": group_id, "docs": {"$push": "$$ROOT"}}})

        return self.collection.aggregate(pipeline, allowDiskUse=allow_disk_use)

    def ensure_index(self, key, unique=False):
        """
        Wrapper for pymongo.Collection.ensure_index for the files collection
        """
        if key in self.files_collection_fields:
            return self._files_collection.create_index(key, unique=unique, background=True)
        else:
            # Store this key to put into metadata collection
            self.meta_keys |= key
            return self._files_collection.create_index("metadata.{}".format(key), unique=unique, background=True)

    def update(self, docs, update_lu=True, key=None):
        """
        Function to update associated MongoStore collection.

        Args:
            docs ([dict]): list of documents
            update_lu (bool) : Updat the last_updated field or not
            key (list or str): list or str of important parameters
        """
        if isinstance(key, str):
            key = [key]
        elif not key:
            key = [self.key]

        key = list(set(key) | self.meta_keys - set(self.files_collection_fields))

        for d in docs:

            search_doc = {k: d[k] for k in key}
            if update_lu:
                d[self.lu_field] = datetime.utcnow()

            metadata = {self.lu_field: d[self.lu_field]}
            metadata.update(search_doc)

            data = json.dumps(jsanitize(d)).encode("UTF-8")
            if self.compression:
                data = zlib.compress(data)
                metadata["compression"] = "zlib"

            self.collection.put(data, metadata=metadata)
            self.transform_criteria(search_doc)

            # Cleans up old gridfs entries
            for fdoc in (self._files_collection.find(search_doc, ["_id"])
                         .sort("uploadDate", -1).skip(1)):
                self.collection.delete(fdoc["_id"])

    def close(self):
        self.collection.database.client.close()



# coding: utf-8
"""
Validator class for document-level validation on Stores. Attach an instance
of a Validator subclass to a Store .schema variable to enable validation on
that Store.
"""

from abc import ABC, abstractmethod
from jsonschema import validate, ValidationError
import pydash

class Validator(ABC):
    """
    A generic class to perform document-level validation on Stores.
    Attach a Validator to a Store during initialization, any all documents
    added to the Store will call .validate_doc() before being added.
    """

    def init(self, strict=False):
        """
        Args:
            strict (bool): Informs Store how to treat Validator: if
            True, will cause build to fail if invalid document
            is found and raise a ValueError, if False will continue
            build but log an error message. In both cases, invalid
            documents will not be stored.
        """
        self.strict = strict

    @abstractmethod
    def is_valid(self, doc):
        """
        Returns (bool): True if document valid, False if document
        invalid
        """
        return NotImplementedError


class StandardValidator(Validator):
    """
    A standard Validator, which allows document validation against a
    provided JSON schema, and also can check that specified keys, if present,
    are MSONable (that is, a Python object can be reconstructed).

    To use, subclass StandardSchema and with your own `schema`
    and `msonable_keypaths` keys.
    """

    @property
    @abstractmethod
    def schema(self):
        """
        Defines a JSON schema for your document,
        which is used by the default `validate_doc()` method.

        This is a standard, with many validators existing, including
        MongoDB's own JSON schema validator (3.6+). Implementing this
        property allows both the document to be validated by a Builder,
        and also makes it possible to enable document-level validation
        inside MongoDB for Mongo-backed Stores.

        Returns (dict): a JSON schema as a Python dict
        """
        return {}

    @property
    @abstractmethod
    def msonable_keypaths(self):
        """
        Optional. Used by the default `validate_doc()` method. Define a
        list of keypaths

        Returns (dict): keys should be the keypath to the relevant value
        in your document, and values should be the relevant class
        """
        return {}

    def is_valid(self, doc):

        # JSON validation
        try:
            validate(doc, schema=self.schema)
            valid_schema = True
        except ValidationError:
            valid_schema = False

        # MSONable validation
        def _validate_doc_msonable(doc):
            """
            For every keypath, will return True if either
            keypath does not exist, or keypath does exist
            and a Python object can be reconstructed.
            Otherwise will return False.
            """

            for keypath, obj in self.msonable_keypaths.items():

                dict_to_check = pydash.get(doc, keypath, None)

                if dict_to_check:
                    try:
                        obj.from_dict(dict_to_check)
                    except:
                        return False

            return True

        valid_msonable = _validate_doc_msonable(doc)

        return valid_schema and valid_msonable




# coding: utf-8
"""
Advanced Stores for behavior outside normal access patterns
"""
import os
import hvac
import json
import boto3
import botocore
import zlib
from datetime import datetime
from maggma.stores import Store, MongoStore
from maggma.utils import lazy_substitute, substitute
from monty.json import jsanitize
from monty.functools import lru_cache


class VaultStore(MongoStore):
    """
    Extends MongoStore to read credentials out of Vault server
    and uses these values to initialize MongoStore instance
    """

    def __init__(self, collection_name, vault_secret_path):
        """
        collection (string): name of mongo collection
        vault_secret_path (string): path on vault server with mongo creds object

        Environment (must be set prior to invocation):
        VAULT_ADDR - URL of vault server (eg. https://matgen8.lbl.gov:8200)
        VAULT_TOKEN or GITHUB_TOKEN - token used to authenticate to vault
        """
        vault_addr = os.getenv("VAULT_ADDR")

        if not vault_addr:
            raise RuntimeError("VAULT_ADDR not set")

        client = hvac.Client(vault_addr)

        # If we have a vault token use this
        token = os.getenv("VAULT_TOKEN")

        # Look for a github token instead
        if not token:
            github_token = os.getenv("GITHUB_TOKEN")

            if github_token:
                client.auth_github(github_token)
            else:
                raise RuntimeError("VAULT_TOKEN or GITHUB_TOKEN not set")
        else:
            client.token = token
            if not client.is_authenticated():
                raise RuntimeError("Bad token")

        # Read the vault secret
        json_db_creds = client.read(vault_secret_path)
        db_creds = json.loads(json_db_creds['data']['value'])

        database = db_creds.get("db")
        host = db_creds.get("host", "localhost")
        port = db_creds.get("port", 27017)
        username = db_creds.get("username", "")
        password = db_creds.get("password", "")

        super(VaultStore, self).__init__(database, collection_name, host, port, username, password)


class AliasingStore(Store):
    """
    Special Store that aliases for the primary accessors
    """

    def __init__(self, store, aliases, **kwargs):
        """
        store (Store): the store to wrap around
        aliases (dict): dict of aliases of the form external key: internal key
        """
        self.store = store
        # Given an external key tells what the internal key is
        self.aliases = aliases
        # Given the internal key tells us what the external key is
        self.reverse_aliases = {v: k for k, v in aliases.items()}
        self.kwargs = kwargs

        kwargs.update({"lu_field": store.lu_field, "lu_type": store.lu_type})
        super(AliasingStore, self).__init__(**kwargs)

    def query(self, properties=None, criteria=None, **kwargs):

        if isinstance(properties, list):
            properties = {p: 1 for p in properties}

        criteria = criteria if criteria else {}
        substitute(properties, self.reverse_aliases)
        lazy_substitute(criteria, self.reverse_aliases)
        for d in self.store.query(properties, criteria, **kwargs):
            substitute(d, self.aliases)
            yield d

    def query_one(self, properties=None, criteria=None, **kwargs):

        if isinstance(properties, list):
            properties = {p: 1 for p in properties}

        criteria = criteria if criteria else {}
        substitute(properties, self.reverse_aliases)
        lazy_substitute(criteria, self.reverse_aliases)
        d = self.store.query_one(properties, criteria, **kwargs)
        substitute(d, self.aliases)
        return d

    def distinct(self, key, criteria=None, **kwargs):
        if isinstance(key, list):
            criteria = criteria if criteria else {}
            # Update to ensure keys are there
            if all_exist:
                criteria.update({k: {"$exists": True} for k in key if k not in criteria})

            results = []
            for d in self.groupby(key, properties=key, criteria=criteria):
                results.append(d["_id"])
            return results

        else:
            criteria = criteria if criteria else {}
            lazy_substitute(criteria, self.reverse_aliases)
            key = self.aliases[key] if key in self.aliases else key
            return self.collection.distinct(key, filter=criteria, **kwargs)

    def groupby(self, keys, properties=None, criteria=None, **kwargs):
        # Convert to a list
        keys = keys if isinstance(keys, list) else [keys]

        # Make the aliasing transformations on keys
        keys = [self.aliases[k] if k in self.aliases else k for k in keys]

        # Update criteria and properties based on aliases
        criteria = criteria if criteria else {}
        substitute(properties, self.reverse_aliases)
        lazy_substitute(criteria, self.reverse_aliases)

        return self.store.groupby(keys=keys, properties=properties, criteria=criteria, **kwargs)

    def update(self, docs, update_lu=True, key=None):
        key = key if key else self.key

        for d in docs:
            substitute(d, self.reverse_aliases)

        if key in self.aliases:
            key = self.aliases[key]

        self.store.update(docs, update_lu=update_lu, key=key)

    def ensure_index(self, key, unique=False, **kwargs):
        if key in self.aliases:
            key = self.aliases
        return self.store.ensure_index(key, unique, **kwargs)

    def close(self):
        self.store.close()

    @property
    def collection(self):
        return self.store.collection

    def connect(self, force_reset=False):
        self.store.connect(force_reset=force_reset)


class SandboxStore(Store):
    """
    Provides a sandboxed view to another store
    """

    def __init__(self, store, sandbox, exclusive=False):
        """
        store (Store): store to wrap sandboxing around
        sandbox (string): the corresponding sandbox
        exclusive (bool): whether to be exclusively in this sandbox or include global items
        """
        self.store = store
        self.sandbox = sandbox
        self.exclusive = exclusive
        super().__init__(key=self.store.key,
                         lu_field=self.store.lu_field,
                         lu_type=self.store.lu_type,
                         validator=self.store.validator)

    @property
    @lru_cache(maxsize=1)
    def sbx_criteria(self):
        if self.exclusive:
            return {"sbxn": self.sandbox}
        else:
            return {"$or": [{"sbxn": {"$in": [self.sandbox]}},
                            {"sbxn": {"$exists": False}}]}

    def query(self, properties=None, criteria=None, **kwargs):
        criteria = dict(**criteria, **self.sbx_criteria) if criteria else self.sbx_criteria
        return self.store.query(properties=properties, criteria=criteria, **kwargs)

    def query_one(self, properties=None, criteria=None, **kwargs):
        criteria = dict(**criteria, **self.sbx_criteria) if criteria else self.sbx_criteria
        return self.store.query_one(properties=properties, criteria=criteria, **kwargs)

    def distinct(self, key, criteria=None, **kwargs):
        criteria = dict(**criteria, **self.sbx_criteria) if criteria else self.sbx_criteria
        return self.store.distinct(key=key, criteria=criteria, **kwargs)

    def groupby(self, keys, properties=None, criteria=None, **kwargs):
        criteria = dict(**criteria, **self.sbx_criteria) if criteria else self.sbx_criteria

        return self.store.groupby(keys=keys, properties=properties, criteria=criteria, **kwargs)

    def update(self, docs, update_lu=True, key=None):
        for d in docs:
            if "sbxn" in d:
                d["sbxn"] = list(set(d["sbxn"] + [self.sandbox]))
            else:
                d["sbxn"] = [self.sandbox]

        self.store.update(docs, update_lu=update_lu, key=key)

    def ensure_index(self, key, unique=False, **kwargs):
        return self.store.ensure_index(key, unique, **kwargs)

    def close(self):
        self.store.close()

    @property
    def collection(self):
        return self.store.collection

    def connect(self, force_reset=False):
        self.store.connect(force_reset=force_reset)


class AmazonS3Store(Store):
    """
    GridFS like storage using Amazon S3 and a regular store for indexing
    Assumes Amazon AWS key and secret key are set in environment or default config file
    """

    def __init__(self, index, bucket, **kwargs):
        """
        Initializes an S3 Store
        Args:
            index (Store): a store to use to index the S3 Bucket
            bucket (str) : name of the bucket
        """
        self.index = index
        self.bucket = bucket
        self.s3 = None
        self.s3_bucket = None
        # Force the key to be the same as the index
        kwargs["key"] = index.key
        super(AmazonS3Store, self).__init__(**kwargs)

    def connect(self, force_reset=False):
        self.index.connect(force_reset=force_reset)
        if not self.s3:
            self.s3 = boto3.resource("s3")
            # TODO: Provide configuration variable to create bucket if not present
            if self.bucket not in self.s3.list_buckets():
                raise Exception("Bucket not present on AWS: {}".format(self.bucket))
            self.s3_bucket = self.s3.Bucket(self.bucket)

    def close(self):
        self.index.close()

    @property
    def collection(self):
        # For now returns the index collection since that is what we would "search" on
        return self.index

    def query(self, properties=None, criteria=None, **kwargs):
        """
        Function that gets data from Amazon S3. This store ignores all
        property projections as its designed for whole document access

        Args:
            properties (list or dict): This will be ignored by the S3
                Store
            criteria (dict): filter for query, matches documents
                against key-value pairs
            **kwargs (kwargs): further kwargs to Collection.find
        """
        for f in self.index.query(criteria=criteria, **kwargs):
            try:
                data = self.s3_bucket.Object(f[self.key]).get()
            except botocore.exceptions.ClientError as e:
                # If a client error is thrown, then check that it was a 404 error.
                # If it was a 404 error, then the object does not exist.
                error_code = int(e.response['Error']['Code'])
                if error_code == 404:
                    self.logger.error("Could not find S3 object {}".format(f[self.key]))
                    break

            if f.get("compression", "") is "zlib":
                data = zlib.decompress(data)

            yield json.loads(data)

    def query_one(self, properties=None, criteria=None, **kwargs):
        """
        Function that gets a single document from Amazon S3. This store
        ignores all property projections as its designed for whole
        document access

        Args:
            properties (list or dict): This will be ignored by the S3
                Store
            criteria (dict): filter for query, matches documents
                against key-value pairs
            **kwargs (kwargs): further kwargs to Collection.find
        """
        f = self.index.query_one(criteria=criteria, **kwargs)
        if f:
            try:
                data = self.s3_bucket.Object(f[self.key]).get()
            except botocore.exceptions.ClientError as e:
                # If a client error is thrown, then check that it was a 404 error.
                # If it was a 404 error, then the object does not exist.
                error_code = int(e.response['Error']['Code'])
                if error_code == 404:
                    self.logger.error("Could not find S3 object {}".format(f[self.key]))
                    return None

            if f.get("compression", "") is "zlib":
                data = zlib.decompress(data)

            return json.loads(data)
        else:
            return None

    def distinct(self, key, criteria=None, all_exist=False, **kwargs):
        """
        Function get to get all distinct values of a certain key in the
        AmazonS3 Store. This searches the index collection for this data

        Args:
            key (mongolike key or list of mongolike keys): key or keys
                for which to find distinct values or sets of values.
            criteria (filter criteria): criteria for filter
            all_exist (bool): whether to ensure all keys in list exist
                in each document, defaults to False
            **kwargs (kwargs): kwargs corresponding to collection.distinct
        """
        # Index is a store so it should have its own distinct function
        return self.index.distinct(key, filter=criteria, **kwargs)

    def groupby(self, keys, properties=None, criteria=None, **kwargs):
        """
        Simple grouping function that will group documents
        by keys. Only searches the index collection

        Args:
            keys (list or string): fields to group documents
            properties (list): properties to return in grouped documents
            criteria (dict): filter for documents to group
            allow_disk_use (bool): whether to allow disk use in aggregation

        Returns:
            command cursor corresponding to grouped documents

            elements of the command cursor have the structure:
            {'_id': {"KEY_1": value_1, "KEY_2": value_2 ...,
             'docs': [list_of_documents corresponding to key values]}

        """
        self.index.groupby(keys, properties, criteria, **kwargs)

    def ensure_index(self, key, unique=False):
        """
        Wrapper for pymongo.Collection.ensure_index for the files collection
        """
        return self.index.ensure_index(key, unique=unique, background=True)

    def update(self, docs, update_lu=True, key=None, compress=False):
        """
        Function to update associated MongoStore collection.

        Args:
            docs ([dict]): list of documents
            key ([str] or str): keys to use to build search doc
            compress (bool): compress the document or not
        """
        now = datetime.now()
        search_docs = []
        for d in docs:
            if isinstance(key, list):
                search_doc = {k: d[k] for k in key}
            elif key:
                search_doc = {key: d[key]}
            else:
                search_doc = {}

            # Always include our main key
            search_doc[self.key] = d[self.key]

            # Remove MongoDB _id from search
            if "_id" in search_doc:
                del search_doc["_id"]

            # Add a timestamp
            if update_lu:
                search_doc[self.lu_field] = now
                d[self.lu_field] = now

            data = json.dumps(jsanitize(d)).encode()

            # Compress with zlib if chosen
            if compress:
                search_doc["compression"] = "zlib"
                data = zlib.compress(data)

            self.s3_bucket.put_object(Key=d[self.key], Body=data, Metadata=search_doc)
            search_docs.append(search_doc)

        # Use store's update to remove key clashes
        self.index.update(search_docs)

    @property
    def last_updated(self):
        return self.index.last_updated

    def lu_filter(self, targets):
        """Creates a MongoDB filter for new documents.

        By "new", we mean documents in this Store that were last updated later
        than any document in targets.

        Args:
            targets (list): A list of Stores

        """
        self.index.lu_filter(targets)

    def __hash__(self):
        return hash((self.index.__hash__, self.bucket))

    def rebuild_index_from_s3_data(self):
        """
        Rebuilds the index Store from the data in S3
        Relies on the index document being stores as the metadata for the file
        """
        index_docs = []
        for file in self.s3_bucket.objects.all():
            # TODO: Transform the data back from strings and remove AWS S3 specific keys
            index_docs.append(file.metadata)

        self.index.update(index_docs)
